package org.apache.maven;

/* ====================================================================
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements.  See the NOTICE file distributed with
 *   this work for additional information regarding copyright ownership.
 *   The ASF licenses this file to You under the Apache License, Version 2.0
 *   (the "License"); you may not use this file except in compliance with
 *   the License.  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

public class ArtifactDownloader
{
    public static final String SNAPSHOT_SIGNATURE = "-SNAPSHOT";

    private File mavenRepoLocal;

    private List remoteRepos;

    private boolean useTimestamp = true;

    private boolean ignoreErrors = false;

    private String proxyHost;

    private String proxyPort;

    private String proxyUserName;

    private String proxyPassword;

    public ArtifactDownloader( String localRepository, List remoteRepositories )
        throws Exception
    {
        setRemoteRepos( remoteRepositories );

        if ( localRepository == null )
        {
            System.err.println( "local repository not specified" );

            System.exit( 1 );
        }

        mavenRepoLocal = new File( localRepository );

        if ( !mavenRepoLocal.exists() )
        {
            if ( !mavenRepoLocal.mkdirs() )
            {
                System.err.println( "Cannot create the specified local repository: " + mavenRepoLocal );

                System.exit( 1 );
            }
        }

        if ( !mavenRepoLocal.canWrite() )
        {
            System.err.println( "Can't write to " + mavenRepoLocal.getAbsolutePath() );

            System.exit( 1 );
        }

        System.out.println( "Using the following for your local repository: " + mavenRepoLocal );
        System.out.println( "Using the following for your remote repositories: " + remoteRepos );
    }

    public File getMavenRepoLocal()
    {
        return mavenRepoLocal;
    }

    private Set downloadedArtifacts = new HashSet();

    public void downloadDependencies( List files )
        throws Exception
    {
        for ( Iterator j = files.iterator(); j.hasNext(); )
        {
            String file = (String) j.next();

            if ( !downloadedArtifacts.contains( file ) )
            {
                File destinationFile = new File( mavenRepoLocal, file );
                // The directory structure for this project may
                // not exists so create it if missing.
                File directory = destinationFile.getParentFile();

                if ( directory.exists() == false )
                {
                    directory.mkdirs();
                }

                if ( destinationFile.exists() && ( file.indexOf( SNAPSHOT_SIGNATURE ) < 0 ) )
                {
                    continue;
                }

                getRemoteArtifact( file, destinationFile );

                if ( !destinationFile.exists() )
                {
                    throw new Exception( "Failed to download " + file );
                }

                downloadedArtifacts.add( file );
            }
        }
    }

    private void setRemoteRepos( List repositories )
    {
        remoteRepos = new ArrayList();

        if ( repositories != null )
        {
            remoteRepos.addAll( repositories );
        }

        if ( repositories.isEmpty() )
        {
            remoteRepos.add( "http://repo1.maven.org" );
        }
    }

    private List getRemoteRepos()
    {
        return remoteRepos;
    }

    private boolean getRemoteArtifact( String file, File destinationFile )
    {
        boolean fileFound = false;

        for ( Iterator i = getRemoteRepos().iterator(); i.hasNext(); )
        {
            String remoteRepo = (String) i.next();

            // The username and password parameters are not being
            // used here. Those are the "" parameters you see below.
            String url = remoteRepo + "/" + file;

            if ( url.startsWith( "file" ) )
            {
                // THe JDK URL for file: should have one or no / instead of // for some reason
                url = replace( url, "file://", "file:" );
            }
            else
            {
                url = replace( url, "//", "/" );
                if ( url.startsWith( "https" ) )
                {
                    url = replace( url, "https:/", "https://" );
                }
                else
                {
                    url = replace( url, "http:/", "http://" );
                }
            }

            // Attempt to retrieve the artifact and set the checksum if retrieval
            // of the checksum file was successful.
            try
            {
                log( "Downloading " + url );
                HttpUtils.getFile( url, destinationFile, ignoreErrors, useTimestamp, proxyHost, proxyPort,
                                   proxyUserName, proxyPassword, true );

                // Artifact was found, continue checking additional remote repos (if any)
                // in case there is a newer version (i.e. snapshots) in another repo
                fileFound = true;
            }
            catch ( FileNotFoundException e )
            {
                log( "Artifact not found at [" + url + "]" );
                // Ignore
            }
            catch ( Exception e )
            {
                // If there are additional remote repos, then ignore exception
                // as artifact may be found in another remote repo. If there
                // are no more remote repos to check and the artifact wasn't found in
                // a previous remote repo, then artifactFound is false indicating
                // that the artifact could not be found in any of the remote repos
                //
                // arguably, we need to give the user better control (another command-
                // line switch perhaps) of what to do in this case? Maven already has
                // a command-line switch to work in offline mode, but what about when
                // one of two or more remote repos is unavailable? There may be multiple
                // remote repos for redundancy, in which case you probably want the build
                // to continue. There may however be multiple remote repos because some
                // artifacts are on one, and some are on another. In this case, you may
                // want the build to break.
                //
                // print a warning, in any case, so user catches on to mistyped
                // hostnames, or other snafus
                log( "Error retrieving artifact from [" + url + "]: " + e );
            }
        }

        return fileFound;
    }

    private String replace( String text, String repl, String with )
    {
        StringBuffer buf = new StringBuffer( text.length() );
        int start = 0, end = 0;
        while ( ( end = text.indexOf( repl, start ) ) != -1 )
        {
            buf.append( text.substring( start, end ) ).append( with );
            start = end + repl.length();
        }
        buf.append( text.substring( start ) );
        return buf.toString();
    }

    private void log( String message )
    {
        System.out.println( message );
    }

    /**
     * Returns the proxyHost.
     *
     * @return Returns the proxyHost.
     */
    protected String getProxyHost()
    {
        return this.proxyHost;
    }

    /**
     * Sets the proxyHost.
     *
     * @param proxyHost The proxyHost to set.
     */
    protected void setProxyHost( String proxyHost )
    {
        this.proxyHost = proxyHost;
    }

    /**
     * Returns the proxyPassword.
     *
     * @return Returns the proxyPassword.
     */
    protected String getProxyPassword()
    {
        return this.proxyPassword;
    }

    /**
     * Sets the proxyPassword.
     *
     * @param proxyPassword The proxyPassword to set.
     */
    protected void setProxyPassword( String proxyPassword )
    {
        this.proxyPassword = proxyPassword;
    }

    /**
     * Returns the proxyPort.
     *
     * @return Returns the proxyPort.
     */
    protected String getProxyPort()
    {
        return this.proxyPort;
    }

    /**
     * Sets the proxyPort.
     *
     * @param proxyPort The proxyPort to set.
     */
    protected void setProxyPort( String proxyPort )
    {
        this.proxyPort = proxyPort;
    }

    /**
     * Returns the proxyUserName.
     *
     * @return Returns the proxyUserName.
     */
    protected String getProxyUserName()
    {
        return this.proxyUserName;
    }

    /**
     * Sets the proxyUserName.
     *
     * @param proxyUserName The proxyUserName to set.
     */
    protected void setProxyUserName( String proxyUserName )
    {
        this.proxyUserName = proxyUserName;
    }
}
