package org.apache.maven;

/* ====================================================================
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements.  See the NOTICE file distributed with
 *   this work for additional information regarding copyright ownership.
 *   The ASF licenses this file to You under the Apache License, Version 2.0
 *   (the "License"); you may not use this file except in compliance with
 *   the License.  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

/**
 * @author <a href="mailto:jvanzyl@apache.org">Jason van Zyl</a>
 * @version $Id: Dependency.java 517014 2007-03-11 21:15:50Z ltheussl $
 */
public class Dependency
{
    /** Dependency Id. */
    private String id;

    /** Version associated with this dependency */
    private String version;

    /** The URL to the dependency's homepage. */
    private String url;

    /** Explictly set JAR name */
    private String jar;

    /** Artifact name */
    private String artifactId;

    /** Group id */
    private String groupId;

    /** The dependency type */
    private String type = "jar";

    /**
     * Default constructor
     */
    public Dependency()
    {
    }

    /**
     * Set the id for this dependency.
     *
     * @param id for this dependency
     */
    public void setId( String id )
    {
        this.id = id;
    }

    /**
     * @return the id attribute (groupId:artifactId)
     */
    public String getId()
    {
        if ( isValid( getGroupId() ) && isValid( getArtifactId() ) )
        {
            // We have something like:
            //
            // <dependency>
            //   <groupId>commons-jelly</groupId>
            //   <artifactId>commons-jelly-tags-velocity</artifactId>
            //   <version>SNAPSHOT</version>
            //  </dependency>

            return getGroupId() + ":" + getArtifactId();
        }

        return id;
    }

    /**
     * Set the group id.
     *
     * @param groupId Group id for the dependency.
     */
    public void setGroupId( String groupId )
    {
        this.groupId = groupId;
    }

    /**
     * Get the group id.
     *
     * @return The group id for the dependency.
     */
    public String getGroupId()
    {
        return groupId;
    }

    /**
     * Get the directory to place the artifact in. If the groupId has been
     * set then use that, otherwise use the id.
     *
     * @return The artifact directory.
     */
    public String getArtifactDirectory()
    {
        if ( isValid( getGroupId() ) )
        {
            return getGroupId();
        }

        return getId();
    }

    /**
     * Get the artifact id.
     *
     * @return The artifact id.
     */
    public String getArtifactId()
    {
        return artifactId;
    }

    /**
     * Set the artifact name.
     *
     * @param artifactId The artifact name of the dependency.
     */
    public void setArtifactId( String artifactId )
    {
        this.artifactId = artifactId;
    }

    /**
     * Gets the artifact name of the dependency. This is always calculated so
     * that the version can be changed dynamically using the maven override
     * facility.
     *
     * @return The artifact name.
     */
    public String getArtifact()
    {
        // If the jar name has been explicty set then use that. This
        // is when the <jar/> element is explicity used in the POM.
        if ( jar != null )
        {
            return jar;
        }

        if ( isValid( getArtifactId() ) )
        {
            return getArtifactId() + "-" + getVersion() + "." + getType();
        }
        else
        {
            return getId() + "-" + getVersion() + "." + getType();
        }
    }

    /**
     * Set the version for this dependency.
     *
     * @param version Version for this dependency
     */
    public void setVersion( String version )
    {
        this.version = version;
    }

    /**
     * Get the version of this dependency.
     *
     * @return Version of this dependency
     */
    public String getVersion()
    {
        return version;
    }

    /**
     * Set the name of the JAR if it cannot be synthesized from the id and
     * version.
     *
     * @param jar Name of the jar
     */
    public void setJar( String jar )
    {
        // This is a check we need because of the jelly interpolation
        // process. If we don't check an empty string will be set and
        // screw up getArtifact() above.
        if ( jar.trim().length() == 0 )
        {
            return;
        }

        this.jar = jar;
    }

    /**
     * Get the name of the JAR. We will attempt to synthesize the name of the
     * JAR from the id and version.
     *
     * @return Name of the jar
     */
    public String getJar()
    {
        return jar;
    }

    /**
     * Set the name of url for the dependency.
     *
     * @param url the dependency's home page
     */
    public void setUrl( String url )
    {
        this.url = url;
    }

    /**
     * Get the name of the url for the dependency.
     *
     * @return Name of the url for the dependency
     */
    public String getUrl()
    {
        return url;
    }

    /**
     * Set the type of the dependency.
     *
     * @return dependency type such as "compile" or "test"
     */
    public String getType()
    {
        return type;
    }

    /**
     * Sets the dependency type such as "compile" or "test"
     *
     * @param type The type of dependency.
     */
    public void setType( String type )
    {
        this.type = type;
    }

    /**
     * Debug string.
     *
     * @return Debugging string.
     */
    public String toString()
    {
        return "Dep[ id:" + getId() + " pid:" + getId() + " ver:" + getVersion() + " ar:" + getArtifact() + " jar:"
            + getJar() + " ]";
    }

    /**
     * Simple check for a value in the POM. Due to the Jelly swizzling
     * fields that aren't set come out as empty strings. This will not
     * be required when the new lazy evaluation mechanism is put in place.
     *
     * @param value POM value to test.
     * @return Is the value valid.
     */
    protected boolean isValid( String value )
    {
        if ( ( value != null ) && !"".equals( value.trim() ) )
        {
            return true;
        }

        return false;
    }
}
