package org.apache.maven.jelly;

/* ====================================================================
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements.  See the NOTICE file distributed with
 *   this work for additional information regarding copyright ownership.
 *   The ASF licenses this file to You under the Apache License, Version 2.0
 *   (the "License"); you may not use this file except in compliance with
 *   the License.  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.jelly.JellyContext;
import org.apache.commons.jelly.TagLibrary;
import org.apache.commons.jelly.XMLOutput;
import org.apache.commons.jelly.expression.Expression;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.maven.MavenConstants;
import org.apache.maven.MavenSession;
import org.apache.maven.jelly.tags.jeez.MavenJeezTagLibrary;
import org.apache.maven.project.Project;
import org.apache.maven.werkz.WerkzProject;
import org.codehaus.plexus.util.StringUtils;

/** Specialized version of a <code>JellyContext</code>.
 *
 * <p>
 * This class simply provides hooks for accessing the jelly
 * run-time execution.  Typical idiom simply includes adding
 * poor-man's-aspects (PMAs) to a <code>JellyContext</code>
 * an explicitly invoking the super-class implementation
 * of a given method.
 * </p>
 *
 * @author <a href="mailto:bob@eng.werken.com">bob mcwhirter</a>
 * @author <a href="mailto:jason@zenplex.com">Jason van Zyl</a>
 *
 * @version $Id: MavenJellyContext.java 517014 2007-03-11 21:15:50Z ltheussl $
 */
public class MavenJellyContext
    extends JellyContext
{
    /** LOGGER for output */
    private final static Log LOGGER = LogFactory.getLog( MavenJellyContext.class );

    // ----------------------------------------------------------------------
    // C O N S T R U C T O R S
    // ----------------------------------------------------------------------

    /** Construct.
     */
    public MavenJellyContext()
    {
        super();
        initializeContext();
    }

    /** Construct.
     *
     *  @param rootContext The root context.
     */
    public MavenJellyContext( URL rootContext )
    {
        super( rootContext );
        initializeContext();
    }

    /** Construct.
     *
     * @param parent Parent MavenJellyContext
     */
    public MavenJellyContext( MavenJellyContext parent )
    {
        super( parent );
        initializeContext();
    }

    /**
     * Create a context with this as the parent
     * @return a new context
     */
    protected JellyContext createChildContext()
    {
        return new MavenJellyContext( this );
    }

    // ----------------------------------------------------------------------
    // I N S T A N C E  M E T H O D S
    // ----------------------------------------------------------------------

    /**
     * Initialize the context with any tag libs or other values we need.
     */
    private void initializeContext()
    {
        MavenJeezTagLibrary jeezTagLib = new MavenJeezTagLibrary();

        registerTagLibrary( "jelly:mavenant", jeezTagLib );
        registerTagLibrary( "", jeezTagLib );

        // Avoid circular dependency
        if ( !isTagLibraryRegistered( "jelly:maven" ) )
        {
            registerTagLibrary( "jelly:maven", "org.apache.maven.jelly.tags.maven.MavenTagLibrary" );
        }
    }

    /** Register a new tag library.
     *
     *  <p>
     *  This implementation calls the superclass
     *  implementation after determining that there
     *  is no other tag-lib registered for the
     *  given <code>uri</code>.
     *  </p>
     *
     *  @param uri Namespace URI to index the taglib.
     *  @param taglib The tag-lib to register.
     */
    public void registerTagLibrary( String uri, TagLibrary taglib )
    {
        if ( !isTagLibraryRegistered( uri ) )
        {
            super.registerTagLibrary( uri, taglib );
        }
    }

    /**
     * Register a new tag library.
     *
     *  <p>
     *  This implementation calls the superclass
     *  implementation after determining that there
     *  is no other tag-lib registered for the
     *  given <code>namespaceURI</code>.
     *  </p>
     *
     *  @param namespaceURI Namespace URI to index the taglib.
     *  @param className The class name of the tag-lib to register.
     */
    public void registerTagLibrary( String namespaceURI, String className )
    {
        if ( !isTagLibraryRegistered( namespaceURI ) )
        {
            super.registerTagLibrary( namespaceURI, className );
        }
    }

    /**
     * Retrieve a registered <code>TagLibrary</code> by namespace URI.
     *
     * @param uri The namespace URI of the tag-library.
     *
     * @return The matching tag-library, or <code>null</code> if none matches.
     */
    public TagLibrary getTagLibrary( String uri )
    {
        return super.getTagLibrary( uri );
    }

    /** Retrieve a variable.
     *
     *  @param name The name of the variable.
     *
     *  @return The value of the variable, if set, otherwise
     *  <code>null</code>.
     */
    public Object getVariable( String name )
    {
        Object value;

        // We need to avoid calling getVariable() on a parent context so
        // that expressions are evaluated lazily
        if ( isInherit() )
        {
            value = super.findVariable( name );
        }
        else
        {
            value = super.getVariable( name );
        }

        if ( value instanceof Expression )
        {
            // TODO: need to handle instances of infinitely recursion when an expression contains itself
            Expression expression = (Expression) value;
            value = expression.evaluate( this );
        }

        return value;
    }

    /**
     * Convert a <code>String</code> property to a
     * <code>Boolean</code> based on its contents.  It would be nice
     * if Jelly would deal with this automatically.
     *
     *  @param key The property key to lookup and convert.
     *
     *  @return The boolean value of the property if convertiable,
     *          otherwise <code>Boolean.FALSE</code>.
     */
    public Boolean getBoolean( String key )
    {
        // After changing the processing everything is now a Boolean already
        // but I want to keep the type check for the time being.

        Object value = getVariable( key );

        if ( value instanceof Boolean )
        {
            return (Boolean) value;
        }

        String stringValue = (String) value;

        if ( "true".equalsIgnoreCase( stringValue ) || "on".equalsIgnoreCase( stringValue ) || "1".equals( value ) )
        {
            return Boolean.TRUE;
        }

        return Boolean.FALSE;
    }

    // ----------------------------------------------------------------------
    // A C C E S S O R S
    // ----------------------------------------------------------------------

    /**
     * Set the mavenSession attribute.
     *
     * @param mavenSession the session for the context
     */
    public void setMavenSession( MavenSession mavenSession )
    {
        setVariable( MavenConstants.SESSION, mavenSession );
    }

    /**
     * Get the mavenSession attribute.
     *
     * @return The
     */
    public MavenSession getMavenSession()
    {
        return (MavenSession) getVariable( MavenConstants.SESSION );
    }

    public void setMavenRepoRemotes( List mavenRepoRemote )
    {
        setVariable( MavenConstants.REPO_REMOTE, convertListToCsvString( mavenRepoRemote ) );
    }

    /**
     * Get the mavenRepoRemote attribute.
     *
     * @return The list of remote repositories.
     */
    public List getMavenRepoRemote()
    {
        // We might have CSV list of remote repositories.
        return convertCsvStringToList( (String) getVariable( MavenConstants.REPO_REMOTE ) );
    }

    /**
     * Convert a CSV list of values into a List.
     *
     * @param csvString CVS list of values.
     * @return The List of value.
     */
    private List convertCsvStringToList( String csvString )
    {
        ArrayList list = new ArrayList();
        String[] s = StringUtils.split( csvString, "," );

        for ( int i = 0; i < s.length; i++ )
        {
            list.add( s[i] );
        }

        return list;
    }

    /**
     * Convert a List into a CSV list of values.
     *
     * @param list list of values.
     * @return The CVS string
     */
    private String convertListToCsvString( List list )
    {
        StringBuffer buf = new StringBuffer();

        for ( int i = 0; i < list.size(); i++ )
        {
            if ( i > 0 )
            {
                buf.append( "," );
            }
            buf.append( list.get( i ) );
        }

        return buf.toString();
    }

    /**
     * Set the location of the local maven repository.
     *
     * @param mavenRepoLocal The local repository.
     */
    public void setMavenRepoLocal( String mavenRepoLocal )
    {
        setVariable( MavenConstants.REPO_LOCAL, mavenRepoLocal );
    }

    /**
     * Get the location of the local maven repository.
     *
     * @return The local repository.
     */
    public String getMavenRepoLocal()
    {
        return (String) getVariable( MavenConstants.REPO_LOCAL );
    }

    /**
     * Set the jelly output sink.
     *
     * @param xmlOutput The jelly output sink.
     */
    public void setXMLOutput( XMLOutput xmlOutput )
    {
        setVariable( MavenConstants.XML_OUTPUT, xmlOutput );
    }

    /**
     * Get the jelly output sink.
     *
     * @return The jelly output sink.
     */
    public XMLOutput getXMLOutput()
    {
        return (XMLOutput) getVariable( MavenConstants.XML_OUTPUT );
    }

    /**
     * Set the online flag.
     *
     * @param online The online flag.
     */
    public void setOnline( Boolean online )
    {
        setVariable( MavenConstants.ONLINE, online );
    }

    /**
     * Get the online flag.
     *
     * @return The online flag.
     */
    public Boolean getOnline()
    {
        return getBoolean( MavenConstants.ONLINE );
    }

    /**
     * Set the maven project.
     *
     * @param project The maven project.
     */
    public void setProject( Project project )
    {
        setVariable( MavenConstants.MAVEN_POM, project );
    }

    /**
     * Get the maven project.
     *
     * @return The maven project.
     */
    public Project getProject()
    {
        return (Project) getVariable( MavenConstants.MAVEN_POM );
    }

    /**
     * Set the proxy host.
     *
     * @param proxyHost The proxy host.
     */
    public void setProxyHost( String proxyHost )
    {
        setVariable( MavenConstants.PROXY_HOST, proxyHost );
    }

    /**
     * Get the proxy host.
     *
     * @return The proxy host.
     */
    public String getProxyHost()
    {
        return (String) getVariable( MavenConstants.PROXY_HOST );
    }

    /**
     * Set the proxy port.
     *
     * @param proxyPort The proxy port.
     */
    public void setProxyPort( String proxyPort )
    {
        setVariable( MavenConstants.PROXY_PORT, proxyPort );
    }

    /**
     * Get the proxy port.
     *
     * @return The proxy port.
     */
    public String getProxyPort()
    {
        return (String) getVariable( MavenConstants.PROXY_PORT );
    }

    /**
     *  Set the proxy user name.
     *
     * @param proxyUserName The user name setting for the proxy.
     */
    public void setProxyUserName( String proxyUserName )
    {
        setVariable( MavenConstants.PROXY_USERNAME, proxyUserName );
    }

    /**
     * Get the proxy user name.
     *
     * @return The proxy user name.
     */
    public String getProxyUserName()
    {
        return (String) getVariable( MavenConstants.PROXY_USERNAME );
    }

    /**
     * Set the proxy password.
     *
     * @param proxyPassword The proxy password.
     */
    public void setProxyPassword( String proxyPassword )
    {
        setVariable( MavenConstants.PROXY_PASSWORD, proxyPassword );
    }

    /**
     * Get the proxy password.
     *
     * @return The ant project.
     */
    public String getProxyPassword()
    {
        return (String) getVariable( MavenConstants.PROXY_PASSWORD );
    }

    //!! Align these these two sets of methods to use getBoolean

    /**
     * Set the debug flag.
     *
     * @param debug The debug flag.
     */
    public void setDebugOn( Boolean debug )
    {
        setVariable( MavenConstants.DEBUG_ON, debug );
    }

    /**
     * Get the debug flag.
     *
     * @return The debug flag.
     */
    public Boolean getDebugOn()
    {
        Boolean b = (Boolean) getVariable( MavenConstants.DEBUG_ON );

        if ( b == null )
        {
            return Boolean.FALSE;
        }

        return b;
    }

    /**
     * Set the emacs mode flag.
     *
     * @param emacsModeOn The emacs mode flag.
     */
    public void setEmacsModeOn( Boolean emacsModeOn )
    {
        setVariable( MavenConstants.EMACS_MODE_ON, emacsModeOn );
    }

    /**
     * Get the emacs mode flag.
     *
     * @return The emacs mode flag.
     */
    public Boolean getEmacsModeOn()
    {
        Boolean b = (Boolean) getVariable( MavenConstants.EMACS_MODE_ON );

        if ( b == null )
        {
            return Boolean.FALSE;
        }

        return b;
    }

    /**
     * Set flag indicating if the remote repository is enabled for use.
     *
     * @param remoteRepositoryEnabled Remote repository usage flag.
     */
    public void setRemoteRepositoryEnabled( Boolean remoteRepositoryEnabled )
    {
        setVariable( MavenConstants.REPO_REMOTE_ENABLED, remoteRepositoryEnabled );
    }

    /**
     * Set flag indicating if the remote repository is enabled for use.
     *
     * @return Remote repository usage flag.
     */
    public Boolean getRemoteRepositoryEnabled()
    {
        return getBoolean( MavenConstants.REPO_REMOTE_ENABLED );
    }

    /**
     * Set the flag indicating the use of the JAR overriding facility.
     *
     * @param mavenJarOverride MavenSession jar override flag.
     */
    public void setMavenJarOverride( Boolean mavenJarOverride )
    {
        setVariable( MavenConstants.JAR_OVERRIDE, mavenJarOverride );
    }

    /**
     * Get the flag indicating the use of the JAR overriding facility.
     *
     * @return MavenSession jar override flag.
     */
    public Boolean getMavenJarOverride()
    {
        return getBoolean( MavenConstants.JAR_OVERRIDE );
    }

    /**
     * Get the maven jar override value for a particular dependency.
     *
     * @param id Id of the project to override.
     * @return MavenSession jar override value for a particular dependency.
     */
    public String getMavenJarOverride( String id )
    {
        return (String) getVariable( MavenConstants.JAR_OVERRIDE_PROPERTY + id );
    }

    /**
     * Get MavenSession home.
     *
     * @return MavenSession home.
     */
    public String getMavenHome()
    {
        return (String) getVariable( MavenConstants.MAVEN_HOME );
    }

    /**
     * retrieve the werkz project object
     * @param werkzProject {@link org.apache.maven.werkz.WerkzProject Werkz Project}
     */
    public void setWerkzProject( WerkzProject werkzProject )
    {
        setVariable( MavenConstants.WERKZ_PROJECT, werkzProject );
    }

    /**
     * retrieve the werkz project object
     * @return the {@link org.apache.maven.werkz.WerkzProject Werkz Project}
     */
    public WerkzProject getWerkzProject()
    {
        return (WerkzProject) getVariable( MavenConstants.WERKZ_PROJECT );
    }

    /**
     * Set the ant project.
     *
     * @param antProject The proxy password.
     */
    public void setAntProject( org.apache.tools.ant.Project antProject )
    {
        setVariable( MavenConstants.MAVEN_ANT_PROJECT, antProject );
    }

    /**
     * Get the ant project.
     *
     * @return The ant project.
     */
    public org.apache.tools.ant.Project getAntProject()
    {
        return (org.apache.tools.ant.Project) getVariable( MavenConstants.MAVEN_ANT_PROJECT );
    }

    /**
     * Display all maven variables
     */
    public void display()
    {
        for ( Iterator i = getVariableNames(); i.hasNext(); )
        {
            String key = (String) i.next();

            if ( key.startsWith( "maven" ) )
            {
                LOGGER.info( key + ": " + getVariable( key ) );
            }
        }
    }

    /**
     * Get local maven home.
     *
     * @return location of Maven local home installation.
     */
    public String getMavenHomeLocal()
    {
        return (String) getVariable( MavenConstants.MAVEN_HOME_LOCAL );
    }

    /**
     * Get plugins location.
     *
     * @return plugins location.
     */
    public String getPluginsDir()
    {
        return (String) getVariable( MavenConstants.MAVEN_PLUGINS_DIR );
    }

    /**
     * Get user plugins location.
     *
     * @return user plugins location.
     */
    public String getUserPluginsDir()
    {
        return (String) getVariable( MavenConstants.MAVEN_USER_PLUGINS_DIR );
    }

    /**
     * Get unpacked plugins location.
     *
     * @return unpacked plugins location.
     */
    public String getUnpackedPluginsDir()
    {
        return (String) getVariable( MavenConstants.MAVEN_UNPACKED_PLUGINS_DIR );
    }

    /**
     * @see JellyContext#setParent(JellyContext)
     */
    public void setParent( JellyContext context )
    {
        super.setParent( context );
    }

    public void resolveRelativePaths( File basedir )
    {
        resolveRelativePath( basedir, MavenConstants.REPO_LOCAL );
        resolveRelativePath( basedir, MavenConstants.MAVEN_USER_PLUGINS_DIR );
        resolveRelativePath( basedir, MavenConstants.MAVEN_UNPACKED_PLUGINS_DIR );
        resolveRelativePath( basedir, MavenConstants.MAVEN_PLUGINS_DIR );
        resolveRelativePath( basedir, MavenConstants.MAVEN_HOME_LOCAL );
        resolveRelativePath( basedir, MavenConstants.MAVEN_HOME );
    }

    private void resolveRelativePath( File basedir, String var )
    {
        String value = (String) getVariable( var );
        File f = new File( value );

        if ( ( f.getParentFile() != null ) && !f.getParentFile().exists() )
        {
            f.getParentFile().mkdirs();
        }

        // Need to check for / at the start to convince windows that it is absolute
        if ( !f.isAbsolute() && !value.startsWith( "/" ) && !value.startsWith( "\\" ) )
        {
            f = new File( basedir, f.getPath() );
            LOGGER.debug( "Resolving " + var + " to " + f );
            setVariable( var, f.getAbsolutePath() );
        }
    }
}
