package org.apache.maven.plugin;

/* ====================================================================
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements.  See the NOTICE file distributed with
 *   this work for additional information regarding copyright ownership.
 *   The ASF licenses this file to You under the Apache License, Version 2.0
 *   (the "License"); you may not use this file except in compliance with
 *   the License.  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.maven.MavenException;
import org.codehaus.plexus.util.StringUtils;
import org.xml.sax.Attributes;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Class to parse plugin XML scripts to find goals and tag libraries.
 *
 * @author Brett Porter <a href="mailto:brett@apache.org">brett@apache.org</a>
 */
final class PluginScriptParser
    extends DefaultHandler
{
    /** The plugin housing. */
    private final JellyScriptHousing jellyScriptHousing;

    /** The goal mapper. */
    private final PluginDefinitionHandler handler;

    /** Logging instance. */
    private static final Log LOGGER = LogFactory.getLog( PluginScriptParser.class );

    /** Storage of defined dynamic tag libraries. */
    private final Set dynaTagLibDecls = new HashSet();

    /**
     * Constructor.
     */
    PluginScriptParser( PluginDefinitionHandler handler, JellyScriptHousing jellyScriptHousing )
    {
        this.handler = handler;
        this.jellyScriptHousing = jellyScriptHousing;
    }

    /**
     * We are looking for namespace declarations like the following:
     *
     * xmlns:doc="doc"
     *
     * Here we know that the given plugin.jelly script (or any jelly script used
     * within Maven) has a dependency on the dyna tag lib named 'doc'. We need to
     * make sure that this dyna tag lib is ready for use when the plugin.jelly
     * script is run.
     *
     * @param prefix Prefix to be used in the jelly script.
     * @param uri Uri of the dyna tag lib.
     * @todo check this works for multiple plugin versions
     */
    public void startPrefixMapping( String prefix, String uri )
    {
        if ( !"".equals( prefix ) && !uri.startsWith( "jelly:" ) && !uri.startsWith( "dummy" ) && !"".equals( uri ) )
        {
            dynaTagLibDecls.add( uri );
            handler.addPluginDynaTagDep( jellyScriptHousing, uri );

            if ( LOGGER.isDebugEnabled() )
            {
                LOGGER.debug( "Caching Taglib Uri --> " + uri );
            }
        }
    }

    /**
     * Handles opening elements of the xml file.
     */
    public void startElement( String uri, String localName, String rawName, Attributes attributes )
    {
        if ( "jelly:core".equals( uri ) && "import".equals( localName ) )
        {
            String importUri = attributes.getValue( "uri" );
            if ( importUri == null )
            {
                importUri = attributes.getValue( "file" );
            }
            /*
            if ( importUri != null )
            {
                // We -could- evaluate this as an expression, but the only thing set at this point is ${basedir}
                // and some expressions will spit out exceptions (eg ${context.getVariable('blah')})
                // TODO: maybe sub in basedir, plugin.resources and plugin.dir?
            }
            */

            if ( importUri == null )
            {
                LOGGER.debug( rawName + " does not have a uri in " + jellyScriptHousing.getName() );
            }
            else
            {
                LOGGER.debug( rawName + " importing from uri " + importUri );
                FileInputStream fis = null;
                try
                {
                    File f = new File( importUri );
                    fis = new FileInputStream( new File( importUri ) );
                    jellyScriptHousing.parse( handler, f.getAbsolutePath(), fis );
                }
                catch ( MavenException e )
                {
                    LOGGER.warn( "Error parsing content from uri " + importUri, e );
                }
                catch ( IOException e )
                {
                    LOGGER.debug( "Skipping: error reading from uri " + importUri );
                }
                finally
                {
                    if ( fis != null )
                    {
                        try
                        {
                            fis.close();
                        }
                        catch ( IOException e )
                        {
                            LOGGER.debug( "WARNING: Cannot close stream!", e );
                        }
                    }
                }
            }
        }
        if ( "project".equals( rawName ) )
        {
            String defaultGoal = attributes.getValue( "default" );

            if ( defaultGoal != null )
            {
                LOGGER
                    .warn( "DEPRECATED: the default goal should be specified in the <build> section of project.xml instead of maven.xml" );
                handler.setDefaultGoalName( defaultGoal );
            }
        }
        else if ( "goal".equals( rawName ) )
        {
            String name = StringUtils.deleteWhitespace( attributes.getValue( "name" ) );
            String prereqs = attributes.getValue( "prereqs" );
            String description = attributes.getValue( "description" );

            handler.addGoal( name, prereqs, description, jellyScriptHousing );
        }
        else if ( "preGoal".equals( rawName ) )
        {
            String name = attributes.getValue( "name" );
            handler.addPreGoal( name, jellyScriptHousing );
        }
        else if ( "postGoal".equals( rawName ) )
        {
            String name = attributes.getValue( "name" );
            handler.addPostGoal( name, jellyScriptHousing );
        }
        /*
        else if ( "attainGoal".equals( rawName ) )
        {
            // Use lazy version
        }
        */
        else if ( "taglib".equals( localName ) && "jelly:define".equals( uri ) )
        {
            String tagLibUri = attributes.getValue( "uri" );

            if ( dynaTagLibDecls.contains( tagLibUri ) )
            {
                handler.removePluginDynaTagDep( jellyScriptHousing, tagLibUri );
            }
            handler.addDynaTagLib( tagLibUri, jellyScriptHousing );
        }
    }

    /**
     * Warning callback.
     *
     * @param spe The parse exception that caused the callback to be invoked.
     */
    public void warning( SAXParseException spe )
    {
        printParseError( "Warning", spe );
    }

    /**
     * Error callback.
     *
     * @param spe The parse exception that caused the callback to be invoked.
     */
    public void error( SAXParseException spe )
    {
        printParseError( "Error", spe );
    }

    /**
     * Fatal error callback.
     *
     * @param spe The parse exception that caused the callback to be invoked.
     */
    public void fatalError( SAXParseException spe )
    {
        printParseError( "Fatal Error", spe );
    }

    /**
     * Print parse error to Sytem.err
     */
    private void printParseError( String type, SAXParseException spe )
    {
        LOGGER
            .error( type + " [line " + spe.getLineNumber() + ", row " + spe.getColumnNumber() + "]: "
                + spe.getMessage() );
    }
}
