package org.apache.maven.repository;

/* ====================================================================
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements.  See the NOTICE file distributed with
 *   this work for additional information regarding copyright ownership.
 *   The ASF licenses this file to You under the Apache License, Version 2.0
 *   (the "License"); you may not use this file except in compliance with
 *   the License.  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.io.File;

import org.apache.maven.project.Dependency;
import org.apache.maven.verifier.ChecksumVerificationException;

/**
 * The interface defining an artifact (which probably lives in the
 * repository).
 *
 * @author <a href="mailto:jason@zenplex.com">Jason van Zyl</a>
 * @version $Id: Artifact.java 517014 2007-03-11 21:15:50Z ltheussl $
 */
public interface Artifact
{
    public static final String OVERRIDE_NONE = null;

    public static final String OVERRIDE_VERSION = "version";

    public static final String OVERRIDE_PATH = "path";

    /**
     * Set the dependency for this JAR artifact.
     *
     * @param dependency Dependency this artifact is based on.
     */
    void setDependency( Dependency dependency );

    /**
     * Get the dependency.
     *
     * @return The dependency this artifact is based on.
     */
    Dependency getDependency();

    /**
     * Set the path to the artifact. The maven jar override facilty can be
     * used to change the path to the artifact if the user specifies the
     * use of a specific version.
     *
     * @param path Path to the artifact.
     */
    void setPath( String path );

    /**
     * Return the path of the artifact using platform specific naming
     * conventions.
     *
     * @return Path to the artifact.
     */
    String getPath();

    /**
     * Generate the path for this artifact given its dependency attributes.
     *
     * @return The generated path of the artifact based on the dependency attributes.
     */
    String generatePath();

    /**
     * Return an URL path that is platform agnostic.
     *
     * @return URL of the artifact.
     */
    String getUrlPath();

    /**
     * Return the url to the checksum file for this artifact.
     *
     * @return URL of the checksum file for this artifact.
     */
    String getChecksumUrl();

    /**
     * Return the name of the artifact.
     *
     * @return Name of the underlying dependency.
     */
    String getName();

    /**
     * Boolean flag indicating whether this artifact is a snapshot.
     *
     * @return Flag indicating whether this artifact is a snapshot.
     */
    boolean isSnapshot();

    /**
     * Boolean flag indicating whether this artifact exists.
     *
     * @return Flag indicating the existance of the artifact in the local repository.
     */
    boolean exists();

    /**
     * Get the location of the artifact in the local file system.
     *
     * @return The location of the artifact in the local file system.
     */
    File getFile();

    /**
     * Verify the artifact.
     *
     * @throws ChecksumVerificationException if verification fails
     */
    void verify()
        throws ChecksumVerificationException;

    public void setOverrideType( String overrideType );

    public String getOverrideType();
    
    public String getDescription();
}
