package org.apache.maven.util;

/* ====================================================================
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements.  See the NOTICE file distributed with
 *   this work for additional information regarding copyright ownership.
 *   The ASF licenses this file to You under the Apache License, Version 2.0
 *   (the "License"); you may not use this file except in compliance with
 *   the License.  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.maven.project.Project;
import org.apache.tools.ant.types.Path;

/**
 * Context/pull tool for use in MavenSession templates.
 *
 * @author <a href="mailto:dlr@collab.net">Daniel Rall</a>
 */
public class MavenTool
{
    /** The maven project the tool is processing */
    private Project project;

    /** list of items to be counted. Key is the object, value is a list
     of objects that have been added that are equal */
    private Map counted = new HashMap();

    /** the counted items in ascending order of number of times counted */
    private Map countedItems = null;

    /** the counted items in descending order of number of times counted */
    private Map countedItemsReversed = null;

    /**
     * Creates a new instance.
     */
    public MavenTool()
    {
        // FIXME: Need a path to the project descriptor.
        //project = MavenUtils.getProject("project.xml");
    }

    /**
     * Accessor for the project property.
     * @return the maven project being processed
     */
    public Project getProject()
    {
        return project;
    }

    /**
     * Resolves a path fragment for the first resource found in
     * <code>directories</code>.
     *
     * <blockquote><code><pre>
     * #set ($pathFragment =
     * $maven.resolvePathFragment($context.toolbox.string.basedir, "foo/bar",
     * ${src.set}))
     * </pre></code></blockquote>
     *
     * @param baseDir The directory to which <code>directories</code>
     * are relative.
     * @param resource The path to the resource to find in
     * <code>directories</code>.
     * @param directories The list of paths to search for
     * <code>resource</code> in..
     * @return The path fragment to <code>resource</code>.
     */
    public String resolvePathFragment( String baseDir, String resource, Path directories )
    {
        return resolvePathFragment( baseDir, resource, directories.list() );
    }

    /**
     * Called by {@link #resolvePathFragment(String, String, Path)}.
     *
     * @see #resolvePathFragment(String, String, Path)
     */
    protected final String resolvePathFragment( String baseDir, String resource, String[] paths )
    {
        String path = null;
        for ( int i = 0; i < paths.length; i++ )
        {
            path = paths[i];
            File d = new File( baseDir, path );
            if ( d.isDirectory() && new File( d, resource ).exists() )
            {
                break;
            }
            else
            {
                // Avoid returning a bogus value on the last iteration.
                path = null;
            }
        }
        if ( path == null )
        {
            System.err.println( "Unable to resolve " + resource + " using a base diretory of " + baseDir );
        }
        return path;
    }

    /**
     * Clear the counter so that there are zero items
     */
    public void resetCount()
    {
        getCounted().clear();
    }

    /**
     * Add an object to the list of counted items.
     * @param object the item to be added.
     */
    public void addToCount( Object object )
    {
        List list = null;
        if ( getCounted().get( object ) == null )
        {
            list = new ArrayList();
        }
        else
        {
            list = (List) getCounted().get( object );
        }
        list.add( object );
        getCounted().put( object, list );
        if ( getCountedItems() != null )
        {
            setCountedItems( null );
            setCountedItemsReversed( null );
        }
    }

    /** Fill the provided sorting map of items from the counted items
     * @param map a map that will sort the objects added to it
     */
    private void fillSortingMap( Map map )
    {
        List list = null;
        Object key = null;
        for ( Iterator index = counted.keySet().iterator(); index.hasNext(); )
        {
            key = index.next();
            list = (List) counted.get( key );
            Integer count = new Integer( list.size() );
            Set items = null;
            if ( map.containsKey( count ) )
            {
                items = (Set) map.get( count );
            }
            else
            {
                items = new TreeSet();
            }
            items.add( key );
            map.put( count, items );
        }
    }

    /**
     * get a list of counted items with the key being the number of times
     * counted (in descending order) and the value any arbitrary item that was
     * counted
     * @return a sorted map of #times counted -> list of objects
     */
    public Map getCountDescending()
    {
        if ( getCountedItemsReversed() == null )
        {
            setCountedItemsReversed( new TreeMap( Collections.reverseOrder() ) );
            fillSortingMap( getCountedItemsReversed() );
        }
        return getCountedItemsReversed();
    }

    /**
     * get a list of counted items with the key being the number of times
     * counted and the value any arbitrary item that was counted
     * @return a sorted map of #times counted -> list of objects
     */
    public Map getCount()
    {
        if ( getCountedItems() == null )
        {
            setCountedItems( new TreeMap() );
            fillSortingMap( getCountedItems() );
        }
        return getCountedItems();
    }

    /** Getter for property counted.
     * @return Value of property counted.
     */
    private Map getCounted()
    {
        return counted;
    }

    /** Getter for property countedItems.
     * @return Value of property countedItems.
     */
    private Map getCountedItems()
    {
        return countedItems;
    }

    /** Setter for property countedItems.
     * @param countedItems New value of property countedItems.
     */
    private void setCountedItems( Map countedItems )
    {
        this.countedItems = countedItems;
    }

    /** Getter for property countedItemsReversed.
     * @return Value of property countedItemsReversed.
     */
    private Map getCountedItemsReversed()
    {
        return countedItemsReversed;
    }

    /** Setter for property countedItemsReverse.
     * @param countedItemsReversed New value of property countedItemsReverse.
     */
    private void setCountedItemsReversed( Map countedItemsReversed )
    {
        this.countedItemsReversed = countedItemsReversed;
    }

    public Integer toInteger( String s )
    {
        return Integer.valueOf( s );
    }
}
