package org.apache.maven.verifier;

/* ====================================================================
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements.  See the NOTICE file distributed with
 *   this work for additional information regarding copyright ownership.
 *   The ASF licenses this file to You under the Apache License, Version 2.0
 *   (the "License"); you may not use this file except in compliance with
 *   the License.  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.io.File;

import org.apache.maven.AbstractMavenComponent;
import org.apache.maven.project.Project;

/**
 * Verify the integrity of the local repository.
 *
 * @author <a href="mailto:jason@zenplex.com">Jason van Zyl</a>
 * @version $Id: LocalSettingsVerifier.java 517014 2007-03-11 21:15:50Z ltheussl $
 */
public class LocalSettingsVerifier
    extends AbstractMavenComponent
{
    /**
     * Default constructor.
     * @param project the project to verify
     */
    public LocalSettingsVerifier( Project project )
    {
        super( project );
    }

    /**
     * Verify user settings before attempting to verify the project.
     *
     * @throws RepoConfigException If an error occurs while trying to verify some basic
     * maven settings.
     */
    public void verifyLocalRepository()
        throws RepoConfigException
    {
        // Get local repo property.
        String localRepoProp = getProject().getContext().getMavenRepoLocal();

        if ( localRepoProp == null )
        {
            // This is not likely to happen any more as it defaults
            // to ${maven.home}/repository which is set in the driver.properties
            throw new RepoConfigException( getMessage( "maven.repo.local.unset" ) );
        }

        File localRepo = new File( localRepoProp );

        if ( !localRepo.exists() )
        {
            System.err.println( getMessage( "directory.nonexistant.warning", localRepo ) );

            if ( !localRepo.mkdirs() )
            {
                throw new RepoConfigException( getMessage( "cannot.create.directory.warning", localRepo ) );
            }
        }

        if ( !localRepo.isDirectory() )
        {
            throw new RepoConfigException( getMessage( "not.directory.warning", localRepo ) );
        }

        if ( !localRepo.canWrite() )
        {
            throw new RepoConfigException( getMessage( "not.writable.warning", localRepo ) );
        }
    }
}
