package org.apache.maven.werkz;

/* ====================================================================
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements.  See the NOTICE file distributed with
 *   this work for additional information regarding copyright ownership.
 *   The ASF licenses this file to You under the Apache License, Version 2.0
 *   (the "License"); you may not use this file except in compliance with
 *   the License.  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

/*
 $Id: WerkzProject.java 517014 2007-03-11 21:15:50Z ltheussl $

 Copyright 2002 (C) The Werken Company. All Rights Reserved.
 
 Redistribution and use of this software and associated documentation
 ("Software"), with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain copyright
 statements and notices.  Redistributions must also contain a
 copy of this document.
 
 2. Redistributions in binary form must reproduce the
 above copyright notice, this list of conditions and the
 following disclaimer in the documentation and/or other
 materials provided with the distribution.
 
 3. The name "werkz" must not be used to endorse or promote
 products derived from this Software without prior written
 permission of The Werken Company.  For written permission,
 please contact bob@werken.com.
 
 4. Products derived from this Software may not be called "werkz"
 nor may "werkz" appear in their names without prior written
 permission of The Werken Company. "werkz" is a registered
 trademark of The Werken Company.
 
 5. Due credit should be given to "the werkz project"
 ( http://werkz.werken.com/ ).
 
 THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.
 
 */

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/** General <code>Goal</code> manager.
 *
 *  <p>
 *  It is not strictly necessary to maintain a <code>Goal</code>
 *  graph, it does assist in organizing and administering a
 *  group of <code>Goal</code>s.
 *
 *  @see Goal
 *
 *  @author <a href="mailto:bob@eng.werken.com">bob mcwhirter</a>
 */
public class WerkzProject
{
    // ------------------------------------------------------------
    //     Instance members
    // ------------------------------------------------------------

    /** Goal index. */
    private Map goals;

    /** Name of the default gual, or null. */
    private String defaultGoalName;

    // ------------------------------------------------------------
    //     Constructors
    // ------------------------------------------------------------

    /** Construct.
     */
    public WerkzProject()
    {
        this.goals = new HashMap();
    }

    // ------------------------------------------------------------
    //     Instance methods
    // ------------------------------------------------------------

    /** Retrieve a valid execution chain to attain the specified goal.
     *
     *  <p>
     *  This method returns one of possibly many different valid
     *  execution chains.  All chains will be of the same length.
     *  This implementation happens to perform a breadth-first
     *  walk, so the chain will grow from the deepest sections
     *  to the showllowest instead of plunging/surging upwards.
     *  <p>
     *
     *  <p>
     *  <pre>
     *        B - D
     *      /       \
     *    A           F
     *      \       /
     *        C - E
     *  </pre>
     *
     *  will result in either FEDCBA or FDEBCA but will
     *  not result in the equally valid FECDBA or FDBECA
     *  sequences.
     *  </p>
     *
     *  <p>
     *  The ordering of the array is such that the index
     *  <code>length - 1</code> (the last index of the array)
     *  points to the <code>Goal</code> specified as the
     *  parameter to the method.
     *  </p>
     *
     *  @param name The name of the goal.
     *
     *  @return The array of goals in the execution chain, terminating
     *          with the requested goal as the last member.
     *
     *  @throws NoSuchGoalException if the <code>name</code> refers to no known goal.
     */

    public Goal[] getExecutionChain( String name )
        throws NoSuchGoalException
    {
        Goal goal = getGoal( name );

        LinkedList chain = new LinkedList();
        LinkedList stack = new LinkedList();

        stack.addLast( goal );

        while ( !stack.isEmpty() )
        {
            goal = (Goal) stack.removeFirst();

            if ( chain.contains( goal ) )
            {
                continue;
            }

            chain.addFirst( goal );

            List precursors = goal.getPrecursors();

            for ( Iterator precursorIter = precursors.iterator(); precursorIter.hasNext(); )
            {
                Goal eachPrecursor = (Goal) precursorIter.next();

                if ( !chain.contains( eachPrecursor ) )
                {
                    stack.addLast( eachPrecursor );
                }
            }

        }

        return (Goal[]) chain.toArray( Goal.EMPTY_ARRAY );
    }

    /** Attempt to attain the specified goal.
     *
     *  @param name The name of the goal to attain.
     *  @param session The context in which to attain goals.
     *
     *  @throws NoSuchGoalException if the <code>name</code> refers to no known goal.
     *  @throws UnattainableGoalException if a precursor or the goal itself had an error.
     *  @throws NoActionDefinitionException if a goal contains no action definition.
     */
    public void attainGoal( String name, Session session )
        throws NoSuchGoalException, UnattainableGoalException, NoActionDefinitionException
    {
        Goal goal = getGoal( name );

        if ( goal == null )
        {
            throw new NoSuchGoalException( name );
        }

        goal.attain( session );
    }

    /** Attempt to percolate the specified goal.
     *
     *  @param name The name of the goal to percolate.
     *  @param session The context in which to percolate goals.
     *
     *  @throws NoSuchGoalException if the <code>name</code> refers to no known goal.
     *  @throws UnattainableGoalException if a postcursor or the goal itself had an error.
     *  @throws NoActionDefinitionException if a goal contains no action definition.
     */
    public void percolateGoal( String name, Session session )
        throws NoSuchGoalException, UnattainableGoalException, NoActionDefinitionException
    {
        Goal goal = getGoal( name );

        if ( goal == null )
        {
            throw new NoSuchGoalException( name );
        }

        goal.percolate( session );
    }

    /** Attempt to attain the specified goal.
     *
     *  @param name The name of the goal to attain.
     *
     *  @throws NoSuchGoalException if the <code>name</code> refers to no known goal.
     *  @throws UnattainableGoalException if a precursor or the goal itself had an error.
     *  @throws NoActionDefinitionException if a goal contains no action definition.
     */
    public void attainGoal( String name )
        throws NoSuchGoalException, UnattainableGoalException, NoActionDefinitionException
    {
        attainGoal( name, new Session() );
    }

    /** Attempt to percolate the specified goal.
     *
     *  @param name The name of the goal to percolate.
     *
     *  @throws NoSuchGoalException if the <code>name</code> refers to no known goal.
     *  @throws UnattainableGoalException if a postcursor or the goal itself had an error.
     *  @throws NoActionDefinitionException if a goal contains no action definition.
     */
    public void percolateGoal( String name )
        throws NoSuchGoalException, UnattainableGoalException, NoActionDefinitionException
    {
        percolateGoal( name, new Session() );
    }

    /** Add a <code>Goal</code> to this manager.
     *
     *  @param goal The <code>Goal</code> to add.
     */
    public void addGoal( Goal goal )
    {
        this.goals.put( goal.getName(), goal );
    }

    /** Retrieve a <code>Goal</code> by name from this manager.
     *
     *  @param name The name of the goal to retrieve.
     *
     *  @return The <code>Goal</code> associated with the <code>name</code>,
     *          or <code>null</code> if there is no such <code>Goal</code>.
     */
    public Goal getGoal( String name )
    {
        return (Goal) this.goals.get( name );
    }

    /** Retrieve a <code>Goal</code> by name from this manager.
     *
     *  @param name The name of the goal to retrieve.
     *  @param create Flag indicating if a new <code>Goal</code> should
     *                be created and returned if none currently exists
     *                bound to <code>name</code>.
     *
     *  @return The existing <code>Goal</code> associated with the <code>name</code>,
     *          or a new <code>Goal</code> if none was previously associated with
     *          <code>name</code>.
     */
    public Goal getGoal( String name, boolean create )
    {
        Goal goal = getGoal( name );

        if ( ( goal == null ) && create )
        {
            goal = new Goal( name );
            addGoal( goal );
        }

        return goal;
    }

    /** Set the default goal name.
     *
     *  @param defaultGoalName The name of the default goal.
     */
    public void setDefaultGoalName( String defaultGoalName )
    {
        this.defaultGoalName = defaultGoalName;
    }

    /** Retrieve the default goal name.
     *
     *  @return The namee of the default goal, or
     *          <code>null</code> if not set.
     */
    public String getDefaultGoalName()
    {
        return this.defaultGoalName;
    }

    /** Retrieve the default goal.
     *
     *  @return The default <code>Goal</code> or
     *          <code>null</code> if a valid default
     *          goal name has not been previously
     *          specified using {@link #setDefaultGoalName}.
     */
    public Goal getDefaultGoal()
    {
        return getGoal( getDefaultGoalName() );
    }

    /** Retrieve an unmodifiable collection of all <code>Goal</code>s 
     *  that are a part of this <code>Project</code>.
     *
     *  @return An unmodifiable collection of all <code>Goal</code>s
     *          that are members of this <code>Project</code>.
     */
    public Collection getGoals()
    {
        return Collections.unmodifiableCollection( this.goals.values() );
    }

    /** Produce a textual representation suitable for debugging.
     *
     *  @return A textual representation suitable for debugging.
     */
    public String toString()
    {
        return "[Werkz: goals=" + this.goals + ";]";
    }

    public void removeGoal( String name )
    {
        goals.remove( name );
    }
}
