package org.apache.maven.plugin;

/* ====================================================================
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements.  See the NOTICE file distributed with
 *   this work for additional information regarding copyright ownership.
 *   The ASF licenses this file to You under the Apache License, Version 2.0
 *   (the "License"); you may not use this file except in compliance with
 *   the License.  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 * ====================================================================
 */

import java.io.File;
import java.io.FileInputStream;

import junit.framework.TestCase;

import org.apache.maven.MavenConstants;
import org.apache.maven.MavenSession;
import org.apache.maven.MavenUtils;
import org.apache.maven.jelly.MavenJellyContext;
import org.apache.maven.project.Project;

import com.werken.forehead.Forehead;

/**
 * @author Brett Porter <a href="mailto:brett@apache.org">brett@apache.org</a>
 *
 */
public class PluginManagerTest
    extends TestCase
{

    /** instance being tested */
    private PluginManager pluginManager;

    /** project to test with */
    private Project project;

    /** directory for fake data */
    private static final String FAKE_BASE = "/target/PluginManagerTest/";

    private static String FS = File.separator;

    /** source base */
    private static final String SOURCE_BASE = FS + "src" + FS + "test" + FS + "java" + FS + "org" + FS + "apache" + FS
        + "maven" + FS + "plugin" + FS;

    /** base directory */
    private String basedir;

    /** fake maven.home */
    private String fakeHome;

    /**
     * Constructor for PluginManagerTest.
     * @param name the name of the test case
     */
    public PluginManagerTest( String name )
    {
        super( name );
    }

    /**
     * Set up the test case
     * @throws Exception when any error occurs
     */
    public void setUp()
        throws Exception
    {
        super.setUp();

        basedir = System.getProperty( "user.dir" );
        assertNotNull( basedir );

        // 1) need to fake out maven home, so that the unpacked plugin dir in the test
        //    directory
        fakeHome = basedir + SOURCE_BASE + "maven.home";
        System.setProperty( "maven.home", fakeHome );

        MavenJellyContext context = MavenUtils.createContext( new File( basedir ) );
        assertNotNull( context );

        // 2) need to fake out maven home local
        String fakeHomeLocal = basedir + FAKE_BASE + "maven.home.local";
        new File( fakeHomeLocal ).mkdirs();
        context.setVariable( MavenConstants.MAVEN_HOME_LOCAL, fakeHomeLocal );

        // 3) Need to fake out unpacked plugins dir so that the plugins are
        //    unpacked locally in the test directory
        String fakeUnpackedPlugins = basedir + FAKE_BASE + "maven.plugin.unpacked.dir";
        new File( fakeUnpackedPlugins ).mkdirs();
        context.setVariable( MavenConstants.MAVEN_UNPACKED_PLUGINS_DIR, fakeUnpackedPlugins );

        // fake forehead up - must be done before MavenSession creation
        FileInputStream fis = new FileInputStream( basedir + SOURCE_BASE + "forehead.conf" );
        Forehead.getInstance().config( fis );
        fis.close();

        // setup maven session
        MavenSession mavenSession = new MavenSession();
        mavenSession.setRootContext( context );
        File projectFile = new File( basedir + "/src/test/touchstone-build/project.xml" );

        // set the project being run
        MavenSession.setRootDescriptorFile( projectFile );

        project = MavenUtils.getProject( projectFile, context );

        // This seems like a glitch in the context assembly that doesn't affect the real world, so work around it
        project.getContext().setMavenRepoLocal( fakeHomeLocal + "/repository" );

        assertNotNull( project );

        pluginManager = mavenSession.getPluginManager();
    }

    /**
     * Delete the fake directory
     * @throws Exception when any error occurs
     * @todo uncomment
     */
    public void tearDown()
        throws Exception
    {
        //        try
        //        {
        //            // clean up directories
        //            String fake = basedir + FAKE_BASE;
        //            FileUtils.deleteDirectory(fake);
        //        }
        //        catch (IOException e)
        //        {
        //            // ignore
        //        }
    }

    /**
     * Make sure the plugin manager has loaded the plugins from the directory specified
     * @throws Exception when any error occurs
     */
    public void testLoading()
        throws Exception
    {
        assertNotNull( pluginManager );
        pluginManager.initialize();
        assertTrue( "clean plugin is not loaded properly", pluginManager.getGoalNames().contains( "clean" ) );
        assertTrue( "clean plugin is not loaded properly", pluginManager.getGoalNames().contains( "clean:clean" ) );
        assertTrue( "clean plugin is not loaded properly", pluginManager.getGoalNames().contains( "clean:original" ) );
    }

    /**
     * Make sure the plugin manager can install a plugin
     * @throws Exception when any error occurs
     */
    public void testInstall()
        throws Exception
    {
        installPlugin( "maven-java-plugin-1.3.jar" );

        assertTrue( "java plugin is not loaded properly", pluginManager.getGoalNames().contains( "java:compile" ) );
    }

    /**
     * Make sure the plugin manager can install a plugin twice without bad
     * effects.
     * @throws Exception when any error occurs
     */
    public void testInstallTwice()
        throws Exception
    {
        installPlugin( "maven-java-plugin-1.3.jar" );

        assertTrue( "java plugin is not loaded properly", pluginManager.getGoalNames().contains( "java:compile" ) );

        installPlugin( "maven-java-plugin-1.3.jar" );

        assertTrue( "java plugin is not loaded properly", pluginManager.getGoalNames().contains( "java:compile" ) );
    }

    /**
     * Make sure the plugin manager can upgrade a plugin
     * @throws Exception when any error occurs
     * @todo what is a valid way to check the correct plugin will be used?
     */
    public void testUpgrade()
        throws Exception
    {
        installPlugin( "maven-clean-plugin-1.2-SNAPSHOT.jar" );

        assertTrue( "upgraded clean plugin is not loaded properly", pluginManager.getGoalNames()
            .contains( "clean:other" ) );
        //assertFalse("original clean plugin is not removed properly",
        //pluginManager.getGoalNames().contains("clean:original"));
    }

    /**
     * Make sure the plugin manager can downgrade a plugin
     * @throws Exception when any error occurs
     * @todo what is a valid way to check the correct plugin will be used?
     */
    public void testDowngrade()
        throws Exception
    {
        installPlugin( "maven-clean-plugin-1.0.jar" );

        assertTrue( "downgraded clean plugin is not loaded properly", pluginManager.getGoalNames()
            .contains( "clean:first" ) );
        //assertFalse("original clean plugin is not removed properly",
        //pluginManager.getGoalNames().contains("clean:original"));
    }

    /**
     * install a plugin
     * @param name the file name of the plugin to install relative to the source base
     * @throws Exception when any error occurs
     */
    private void installPlugin( String name )
        throws Exception
    {
        assertNotNull( pluginManager );
        pluginManager.initialize();
        File pluginToInstall = new File( basedir + SOURCE_BASE + name );
        pluginManager.installPlugin( pluginToInstall, project );
    }

    /**
     * delete a plugin
     * @param name the file name of the plugin to install relative to the source base
     * @throws Exception when any error occurs
     */
    private void deletePlugin( String name )
        throws Exception
    {
        File plugin = new File( fakeHome + "/plugins/" + name );
        if ( plugin.exists() )
        {
            plugin.delete();
        }
    }

    /**
     *  Test extracted plugins are loaded from maven.home - for the bootstrap.
     *  @throws Exception on failure
     */
    public void testExtractedPluginInMavenHome()
        throws Exception
    {
        assertNotNull( pluginManager );
        pluginManager.initialize();

        assertTrue( "check extracted jar plugin is loaded", pluginManager.getGoalNames().contains( "jar:jar" ) );
    }

    /**
     * Test plugin property substitution.
     */
    public void testPluginPropertySubstitution()
        throws Exception
    {
        assertNotNull( pluginManager );
        pluginManager.initialize();
        pluginManager.setupBaseContext( project );

        MavenJellyContext context = pluginManager.getPluginContext( "maven-jar-plugin" );
        assertNotNull( "check context exists", context );

        assertEquals( "check plugin.resources", fakeHome + FS + "plugins" + FS + "jar" + FS + "plugin-resources",
                      ( (File) context.getVariable( "plugin.resources" ) ).getAbsolutePath() );
        assertEquals( "check plugin property",
                      fakeHome + FS + "plugins" + FS + "jar" + FS + "plugin-resources/foo.bar", context
                          .getVariable( "maven.jar.foo" ) );
    }
}
